import { FC, ReactElement, useEffect, useState } from 'react'
import { trigger, on, off, MyEvent } from 'sunny-js/util/DOMEvent'
import { noop } from 'sunny-js/util/function'

export interface ICSSModuleStyle { [className: string]: string }
interface IThemeFileLoader {
  (): Promise<{ default: ICSSModuleStyle }>
  loaded?: boolean | undefined
}
// 用来存放等待加载的，主题样式文件加载器
interface IAddedThemes { [theme: string]: { [fn: string]: IThemeFileLoader } | IThemeFileLoader }
const addedThemes: IAddedThemes = {}

interface ILoaderOptions {
  onLoading?: () => void
  onSuccess?: () => void
  onError?: () => void
  onComplete?: () => void
}
function themeFileLoader(
  {
    onLoading = noop,
    onSuccess = noop,
    onError = noop,
    onComplete = noop,
    theme = currentTheme(),
    themes = addedThemes,
  }: ILoaderOptions & { theme?: string, themes?: IAddedThemes } = {}
) {
  let loaders: IThemeFileLoader[] = []
  const pickedThemes = themes[theme]
  if (typeof pickedThemes === 'function') {
    loaders.push(pickedThemes)
  } else {
    loaders = Object.values(pickedThemes || {})
  }

  // 打开loading
  onLoading()
  Promise.all(loaders.map(promiseLoader => promiseLoader().then(ESModule => {
    // 标记为已加载
    promiseLoader.loaded = true
    // 已完成的主题加载器，下次不再需要使用了
    clearThemeLoaders(theme, promiseLoader)
    return ESModule
  })))
    .then(onSuccess)
    .catch(onError)
    .finally(onComplete)
}

function pickThemeLoaders(theme = currentTheme()) {
  return addedThemes[theme] = addedThemes[theme] || {}
}

function clearThemeLoaders(theme: string, loader?: IThemeFileLoader) {
  const loaders = pickThemeLoaders(theme)
  if (loader) {
    delete loaders[loader.toString()]
  } else {
    delete addedThemes[theme]
  }
}

interface IAsyncThemes { [theme: string]: IThemeFileLoader }
/**
 * 为当前组件添加主题相关样式文件
 * @example
 * // 多样式文件用例
 * addTheme({
 *   light: () => import('./style.light.scss'),
 *   dark: () => import('./style.dark.scss')
 * })
 */
export function addTheme (
  themes: IAsyncThemes,
  options?: ILoaderOptions) {
  // 往addedThemes增加每个样式文件的loader
  Object.keys(themes).forEach(theme => {
    const loaders = pickThemeLoaders(theme)
    const loader = (themes as IAsyncThemes)[theme]
    const key = loader.toString()
    const savedLoader = loaders[key]
    // 没有被加载的模块，且没有往addedThemes存过，则添加
    if (!loader.loaded && !savedLoader) {
      loaders[key] = loader
    }
  })
  // 主动加载当前样式文件
  themeFileLoader({
    ...options,
    themes
  })
}

// 获取当前主题
export function currentTheme () {
  return document.documentElement.getAttribute('data-theme')
}

// 主题切换事件名称
const ChangeEvent = 'theme:changed'
export interface IEventData {
  theme: string
}
/**
 * 切换主题
 */
export function changeTheme (
  theme: string,
  {
    onSuccess = noop,
    ...options
  }: ILoaderOptions = {}
) {
  const setTheme = () => document.documentElement.setAttribute('data-theme', theme)
  const _currentTheme = currentTheme()
  if (_currentTheme === theme) {
    return
  }

  /**
   * 首次设置场景，可以进行快速设置，不需要关心文件异步加载任务。
   * 调用addTheme函数的都会主动加载主题样式，如果theme为空，则会进入addedThemes加载队列
   */
  if (!_currentTheme) {
    setTheme()
  }

  /**
   * 非首次设置场景，需要等待样式文件都加在完毕后，再调用setTheme，
   * 避免可能出现的空白现象，使主题平滑切换
   * 同时也有主动完成addedTheme加载队列的作用
   */
  themeFileLoader({
    ...options,
    theme, // 加载指定主题
    onSuccess () {
      // 第一步，先设置主题，其他逻辑快就可以通过currentTheme访问更改后的主题
      setTheme()
      // 第二步，触发主题切换事件，其他逻辑块可能会从loader加载器中读取css module
      trigger<IEventData>(ChangeEvent, { theme })
      onSuccess()
    }
  })
}

/**
 * 主题切换工具组件
 * 不作为功能的主要部分，为你简化主题切换按钮提供支持
 * css module theming https://github.com/css-modules/css-modules/blob/master/docs/theming.md#theming
 */
type ThemeSwitchProps = {
  children: ({}: {
    change: (theme: string) => void,
    // theme: string
    loading: boolean
  }) => ReactElement,
  // theme: string,
  onChange?: (theme: string) => void
} & ILoaderOptions
export const ThemeSwitch: FC<ThemeSwitchProps> = ({
  children,
  // theme: defaultTheme,
  onChange = noop,
  onLoading = noop,
  onSuccess = noop,
  onError = noop,
  onComplete = noop,
}) => {
  const [loading, setLoading] = useState(false)
  const change = theme => changeTheme(theme, {
    onLoading() {
      setLoading(true)
      onLoading()
    },
    onSuccess() {
      onChange(theme)
      onSuccess()
    },
    onComplete() {
      setLoading(false)
      onComplete()
    },
    onError
  })
  return children({
    loading,
    change,
  })
}

interface IThemePicker {
  themes: IAsyncThemes,
  children?: ({}: ReturnType<typeof useThemePicker>) => ReactElement,
}
export const ThemePicker: FC<IThemePicker> = ({ children, themes }) => {
  return children(useThemePicker(themes))
}
/**
 * useThemePicker hook
 * 跟随主题切换事件，进行CSS Module样式对象切换
 */
export function useThemePicker (themes: IAsyncThemes) {
  const [result, setResult] = useState<{
    style: ICSSModuleStyle
    loading: boolean
    error: boolean
  }>({
    style: null,
    loading: false,
    error: false
  })

  useEffect(() => {
    // 添加到队列中
    addTheme(themes)
    // 样式加载完成后，且触发了主题切换事件
    const _loadTheme: { (): void, _cachedStyle? } = () => {
      // 加载当前主题样式
      const promiseLoader = themes[currentTheme()]
      promiseLoader && promiseLoader()
        .then(({ default: _style }) => {
          // 跳过不必要的调用
          if (!Object.is(_loadTheme._cachedStyle, _style) ) {
            result.style = _style
            result.error = false
            setResult({ ...result })
          }
          _loadTheme._cachedStyle = _style
        })
        .catch(() => {
          result.error = true
          setResult({ ...result })
        })
    }
    // 初始化时主动加载一次
    _loadTheme()
    /**
     * 随后响应主题切换事件。
     * 主题切换前，changeTheme方法会主动等待样式文件加载完后触发主题切换事件
     */
    on(ChangeEvent, _loadTheme)
    return () => off(ChangeEvent, _loadTheme)
  }, [])

  result.style = result.style || {}
  result.loading = !result.style && !result.error
  return result
}

interface IThemeProvider {
  children: (theme: string) => ReactElement,
}
export const ThemeProvider: FC<IThemeProvider> = ({ children }) => {
  return children(useThemeProvider())
}
/**
 * useThemeProvider hook
 * 提供react hook api风格访问currentTheme()
 */
export function useThemeProvider () {
  const [theme, setTheme] = useState<string>(currentTheme())
  useEffect(() => {
    const _onChange = (e: MyEvent<IEventData>) => setTheme(e.data.theme)
    on(ChangeEvent, _onChange)
    return () => off(ChangeEvent, _onChange)
  }, [])
  return theme
}

/**
 * 监听主题切换事件
 */
export function onThemeChange(callback: (e: MyEvent<IEventData>) => void, options?: boolean | AddEventListenerOptions) {
  on(ChangeEvent, callback, options)
}
/**
 * 卸载主题切换事件
 */
export function offThemeChange(callback: (e: MyEvent<IEventData>) => void, options?: boolean | AddEventListenerOptions) {
  off(ChangeEvent, callback, options)
}
